/* Copyright 2020 Michael Thies <mail@mhthies.de>
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */

body.pushable>.pusher, body:not(.pushable) {
    background: url("prism.png");
    padding-top: 60px;
}

.shc.label{
    font-weight: bold;
}
.shc.label.float-btn {
    line-height: 36px;
}
.shc.slider-label-container:after{
    clear: both;
    display: block;
    content: "";
}
.shc.slider-label-container{
    margin-bottom: 8px;
}
.shc.slider-label-container>.shc.label{
    line-height: 26px;
}


/* ************************** *
 * Semantic UI extensions     *
 * ************************** */

/* The following responsive features for Semantic UI are written by GitHub Users @sgeraldes and
 * @aliok <aliok@redhat.com>.
 * Source: https://github.com/Semantic-Org/Semantic-UI/issues/1114#issuecomment-201057438 */

/* Semantic UI has these classes, however they're only applicable to
 * grids, containers, rows and columns.
 * plus, there isn't any `mobile hidden`, `X hidden` class.
 * this snippet is using the same class names and same approach
 * plus a bit more but to all elements.
 *
 * see https://github.com/Semantic-Org/Semantic-UI/issues/1114
 */

/* Mobile */
@media only screen and (max-width: 767px) {
  [class*="mobile hidden"],
  [class*="tablet only"]:not(.mobile),
  [class*="computer only"]:not(.mobile),
  [class*="large screen only"]:not(.mobile),
  [class*="widescreen only"]:not(.mobile),
  [class*="or lower hidden"] {
    display: none !important;
  }
}

/* Tablet / iPad Portrait */
@media only screen and (min-width: 768px) and (max-width: 991px) {
  [class*="mobile only"]:not(.tablet),
  [class*="tablet hidden"],
  [class*="computer only"]:not(.tablet),
  [class*="large screen only"]:not(.tablet),
  [class*="widescreen only"]:not(.tablet),
  [class*="or lower hidden"]:not(.mobile) {
    display: none !important;
  }
}

/* Computer / Desktop / iPad Landscape */
@media only screen and (min-width: 992px) and (max-width: 1199px) {
  [class*="mobile only"]:not(.computer),
  [class*="tablet only"]:not(.computer),
  [class*="computer hidden"],
  [class*="large screen only"]:not(.computer),
  [class*="widescreen only"]:not(.computer),
  [class*="or lower hidden"]:not(.tablet):not(.mobile) {
    display: none !important;
  }
}

/* Large Monitor */
@media only screen and (min-width: 1200px) and (max-width: 1919px) {
  [class*="mobile only"]:not([class*="large screen"]),
  [class*="tablet only"]:not([class*="large screen"]),
  [class*="computer only"]:not([class*="large screen"]),
  [class*="large screen hidden"],
  [class*="widescreen only"]:not([class*="large screen"]),
  [class*="or lower hidden"]:not(.computer):not(.tablet):not(.mobile) {
    display: none !important;
  }
}

/* Widescreen Monitor */
@media only screen and (min-width: 1920px) {
  [class*="mobile only"]:not([class*="widescreen"]),
  [class*="tablet only"]:not([class*="widescreen"]),
  [class*="computer only"]:not([class*="widescreen"]),
  [class*="large screen only"]:not([class*="widescreen"]),
  [class*="widescreen hidden"],
  [class*="widescreen or lower hidden"] {
    display: none !important;
  }
}

/* More colors for toggle checkbox */
.ui.toggle.checkbox.red input:checked~label:before {
    background-color: #db2828!important;
}
.ui.toggle.checkbox.orange input:checked~label:before {
    background-color: #f2711c!important;
}
.ui.toggle.checkbox.yellow input:checked~label:before {
    background-color: #fbbd08!important;
}
.ui.toggle.checkbox.olive input:checked~label:before {
    background-color: #b5cc18!important;
}
.ui.toggle.checkbox.green input:checked~label:before {
    background-color: #21ba45!important;
}
.ui.toggle.checkbox.teal input:checked~label:before {
    background-color: #00b5ad!important;
}
.ui.toggle.checkbox.violet input:checked~label:before {
    background-color: #6435c9!important;
}
.ui.toggle.checkbox.purple input:checked~label:before {
    background-color: #a333c8!important;
}
.ui.toggle.checkbox.pink input:checked~label:before {
    background-color: #e03997!important;
}
.ui.toggle.checkbox.brown input:checked~label:before {
    background-color: #a5673f!important;
}
.ui.toggle.checkbox.grey input:checked~label:before {
    background-color: #767676!important;
}
.ui.toggle.checkbox.black input:checked~label:before {
    background-color: #1b1c1d!important;
}

/* Fix divided lists in divided lists and segments */
.ui.divided.list .item .divided.list>.item {
    border-top: 1px solid rgba(34,36,38,.15);
}
.ui.divided.list .item .divided.list>.item:first-child {
    border-top: none;
}
.segment>.ui.divided.list:first-child {
    padding: 0;
}


/* ************************** *
 * Widget things              *
 * ************************** */
.IroWheel, .IroSlider {
  margin-left: auto;
  margin-right: auto;
}

.ui.list>.item [class*="right floated"].shc.hiderow-button {
    margin: -7px -10px -7px 10px;
}

.shc.image-container {
    position: relative;
    max-width: 100%;
    margin: 0 auto;
}
.shc.image-container>.background {
    width: 100%;
}
.shc.image-container>.item {
    position: absolute;
    transform: translateX(-50%) translateY(-50%);
}
.shc.image-container>.item.with-popup>* {
    cursor: pointer;
}
