{% from "buttons.inc.htm" import render_button %}
<label class="shc label float-btn">{{ label }}</label>
<div class="ui buttons right floated">
    {% for button in buttons %}
        {{ render_button(button) }}
    {% endfor %}
</div>
