<div class="shc slider-label-container">
    <label class="shc label">{{ label }}</label>
    <div class="ui right floated {{ color or "black" }} label value-display"><i class="fitted notched circle loading icon"></i></div>
</div>
<div class="ui {{ color }} labeled slider" id="slider-1" data-widget="slider" data-id="{{ id }}"></div>
